################################################################################
#                                                                              #
#   calnexRest.tcl                                                             #
#                                                                              #
#   TCL interface to the new line of Calnex products                           #
#                                                                              #
#   Copyright (c) Calnex Solutions Ltd 2008 - 2019                             #
#                                                                              # 
#   The contents of this file should not be copied or distributed              #
#   without permission being granted by Calnex Solutions Ltd.                  #
#                                                                              #
#   All rights reserved.                                                       #
#                                                                              #
################################################################################
package require rest
package require json
package require json::write

# set [namespace current]::__IpAddress ""
# set [namespace current]::__lastError ""
variable __IpAddress ""
variable __lastError ""


proc __chkForError { command } {
	# variable [namespace current]::__lastError
	variable __lastError
    if {[string compare $__lastError ""] != 0 } {
		logDebug "$__lastError (command=\"$command\")"
		error "$__lastError (command=\"$command\")"
	}
}

proc pxInit {IpAddr} {
	# global [namespace current]::__IpAddress
	# global [namespace current]::__lastError
	variable __IpAddress
	variable __lastError
	
	set __lastErr ""
	if {$IpAddr == ""} {
		set __lastError "Must specify an IP Address for the instrument"
	} else {
		set __IpAddress $IpAddr
	}	
}
proc calnexInit {IpAddr} {
	# global [namespace current]::__IpAddress
	# global [namespace current]::__lastError
	variable __IpAddress
	variable __lastError
	
	set __lastErr ""
	if {$IpAddr == ""} {
		set __lastError "Must specify an IP Address for the instrument"
	} else {
		set __IpAddress $IpAddr
		set model [calnexGetVal instrument/information HwType]
		set sn [calnexGetVal instrument/information SerialNumber]
		puts "$model $sn"
	}
    __chkForError "calnexInit"

}

proc calnexSendReq {req url body} {
	# variable [namespace current]::__IpAddress
	# global [namespace current]::__lastError
	variable __IpAddress
	variable __lastError

	set httpErrorDict [dict create 400 "Bad Request" 404 "Not Found" 500 "Internal Server Error" 501 "Not Implemented"]
	set ret ""
    set __lastError ""
    if {$__IpAddress == ""} {
        set __lastError "IP address not configured - call calnexInit before any other calls"
        set ret ""
    } else {
		set config "method ${req} headers \{Content-Type application/json\}"
		#set ret [rest::simple "http://$__IpAddress/api/${url}?format=json" "" $config $body]
		
		if { [catch {
			set ret [rest::simple "http://$__IpAddress/api/${url}?format=json" "" $config $body]
		} errmsg ] } {
			set httpErrorString ""
			set httpStr [lindex [split $errmsg] 0]
			if {$httpStr == "HTTP"} {
				set httpErrorNum [lindex [split $errmsg] 1]
				puts $errmsg
				puts $httpErrorNum
				if {[dict exists $httpErrorDict $httpErrorNum]} {
					set httpErrorString [dict get $httpErrorDict $httpErrorNum]
				}
				#throw $httpErrorNum "HTTP $httpErrorNum: $httpErrorString ($url)"
				set __lastError "HTTP $httpErrorNum: $httpErrorString"
			} else {
				set __lastError "$errmsg"
			}
		}

	}
    return $ret
}

proc calnexGet {url args} {
    set body [json::write object {*}$args]
	set res [json::json2dict [calnexSendReq get $url $body]]
    __chkForError "calnexGet $url"
	return $res
}

proc calnexGetVal {url arg} {
	# global [namespace current]::__lastError
	variable __lastError

    set body ""
	set ret ""
	set calnexSendReqRet [calnexSendReq get $url $body]
    __chkForError "calnexGetVal $url"
	set res [json::json2dict $calnexSendReqRet]
	if {[dict exists $res $arg]} {
		set ret [dict get $res $arg]
	} else {
		set __lastError "$arg does not exist in response: $res"
	}
    __chkForError "calnexGetVal $url"
	return $ret
}

proc calnexSet {url args} {
	set body [json::write object {*}$args]
	calnexSendReq put $url $body
    __chkForError "calnexSet $url"
}

proc calnexCreate {url args} {
	set body [json::write object {*}$args]
	calnexSendReq post $url $body
    __chkForError "calnexCreate $url"
}

proc calnexDel {url} {
	calnexSendReq delete $url ""
    __chkForError "calnexDel $url"
}

#
# Old syntax - kept for backwards compatibility
#

proc p100req {req url body} {
	return [calnexSendReq $req $url $body]
}

proc p100get {url args} {
	return [calnexGet $url {*}$args]
}

proc p100set {url args} {
	calnexSet $url {*}$args
}

proc p100create {url args} {
	calnexCreate $url {*}$args
}

proc p100del {url} {
	calnexDel $url
}

proc a100req {req url body} {
	return [calnexSendReq $req $url $body]
}

proc a100get {url args} {
	return [calnexGet $url {*}$args]
}

proc a100set {url args} {
	calnexSet $url {*}$args
}

proc a100create {url args} {
	calnexCreate $url {*}$args
}

proc a100del {url} {
	calnexDel $url
}

#
# Helpers
#
global _ConfigBin
global _ConfigJson
set _ConfigBin "method {get} headers \{Content-Type application/octet-stream\}"
set _ConfigJson "method {get} headers \{Content-Type application/json\}"


#-------------------------------------------------------------------------------
# downLoadFile --
#	Download a file from the instrument
# Arguments:
#	folderType	If "SessionsFolder" or "ReportFolder" then binary data;
#				assume text otherwise
#	srcFolder	The name of the folder on the instrument. For sessions files
#				this is the name of the session folder e.g. Session_<date>
#	file 		The name of the file - for capture files, this is the name of
#				the file in the Session folder
#	destFolder	The name of the folder on the local machines where the remote
#				folder will be saved
# Results:
#	Raises an error if the file cannot be found on the isntrument
#	If the local file or folder can't be accessed then tcl will raise a file
#	access error
#
proc downloadFile {folderType srcFolder filename destFolder} {
	variable _ConfigBin
	variable _ConfigJson
	# variable [namespace current]::__IpAddress
	variable __IpAddress

	
	set remoteFile "$srcFolder/$filename"
	set url "cat/filecommander/download/$folderType?AsAttachment=true&FileId=$remoteFile"

	set f [file join "$destFolder" "$filename"]
	set fid [open $f "w"]
	if {$folderType == "SessionsFolder" || $folderType == "ReportFolder"} {
		fconfigure $fid -translation binary
		set config $_ConfigBin
	} else {
		set config $_ConfigJson			
	}
		
	set err [catch {
		puts $fid [rest::simple "http://$__IpAddress/api/${url}" "" $config ""]
	} res]
	#puts $err

	if {$err != 0} {
		error "Unable to download $filename from '$srcFolder'. Please check that this folder and file exist."
	}
		
	close $fid
	
	return
}
	
#-------------------------------------------------------------------------------
# downloadSessionFile --
#	Download a capture file from the instrument
# Arguments:
#	srcFolder	The name of the Session folder e.g. Session_<data>
#	file		The name of the capture in the Session folder
#	destFolder	The name of the folder on the local PC where the Session folder
#				will be saved
# Results:
#	None
#
proc downloadSessionFile {srcFolder file destFolder} {
	set folderType "SessionsFolder"
	downloadFile $folderType $srcFolder $file $destFolder
	return
}

#-------------------------------------------------------------------------------	
# downloadSessionFiles --
#	Download all the captures files in a Session folder
# Arguments:
#	srcFolder	The name of the Session folder on the instrument
#	destFolder	The name of the folder on the local PC where the Session folder
#				will be saved
# Results:
#	Raises an error if the Session folder on the instrument can't be accessed.
#
proc downloadSessionFiles {srcFolder destFolder} {
	variable ConfigJson
	# variable [namespace current]::__IpAddress
	variable __IpAddress

	set url "cat/filecommander/SessionsFolder?FileId=$srcFolder&format=json"

	set err [catch {
		set resp [rest::simple "http://$__IpAddress/api/${url}" "" $ConfigJson ""]			
	} res]
	#puts $err
		
	if {$err != 0} {
		error "Unable to access $srcFolder. Please check that this folder exists."
	}
		
	set files [dict get [json::json2dict $resp] ChildrenElements]
	#puts $files
	foreach file $files {
		downloadSessionFile $srcFolder $file $destFolder			
	}
}

#-------------------------------------------------------------------------------	
# calnexCatDone --
#	Wait for the CAT to finish opening files and processing data
# Arguments:
#	None
# Results:
#	Returns True when the CAT no longer indicates that it is processing or opening.
# 	The polling period is 1 second with a meximum of 60 re-tries. If the re-try
#	count is exceeded, False is returned.
#
proc calnexCatDone {} {
	set catStatus [calnexGet "/cat/general/status"]
	set catCurrentlyProcessing [dict get $catStatus "IsApiCurrentlyProcessing"]
	set catOpening [dict get $catStatus "IsOpeningInProgress"]
	set retry 0
	
    #puts "calnexCatDone: processing - $catCurrentlyProcessing; opening: $catOpening"
	while {$catCurrentlyProcessing || $catOpening} {
		after 1000
		set catStatus [calnexGet "/cat/general/status"]
		set catCurrentlyProcessing [dict get $catStatus "IsApiCurrentlyProcessing"]
		set catOpening [dict get $catStatus "IsOpeningInProgress"]
		#puts "cp: $catCurrentlyProcessing o: $catOpening"
		incr retry
		if {$retry > 60} {
			return False
		}
        #puts "Retry--------------> $catCurrentlyProcessing / $catOpening"
	}
	return True
}

# calnexCatGenerateReport --
#	Generate a report in the CAT and then (optionally) download it to the local PC
# Arguments:
#	reportName	The name of the report to be generated
#	withCharts	If TRUE, then charts will be included in the report.
#	destFolder	(Optional) The name of the folder on the local PC where the report
#				will be saved. The path to the folder will be created if required.
#				If destFolder is not specified then the report will be saved in the
#				current working directory (i.e. where the script is executing)
# Returns:
#	None
proc calnexCatGenerateReport {withCharts reportName {destFolder "./"}} {
	if {[calnexCatDone]} {
		calnexGet "/cat/report/data"
		calnexCreate "/cat/report" RenderCharts ${withCharts} ReportFilename ${reportName}
	
		# Report is now generated. Download it...
		downloadFile "ReportFolder" "." $reportName $destFolder
		
		return
	} else {
		return -code error "CAT still processing"
	}
}

proc calnexCatGet1ppsMissingCount {} {
	set retval 0
	set resp [calnexGetVal "/cat/measurement/1ppsTEAbsolute/-/ONEPPS/-/series/Errors" "Plots"]
	set plots [llength $resp]
	if {$plots > 0} {
		set pts [lindex $resp 0]
		set points [dict get $pts "Points"]
		set retval [llength $points]
	}
	return $retval
}
